/* Copyright (C) 2000-2006  Thomas Bopp, Thorsten Hampel, Ludger Merkens, Robert Hinn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ldap.pike,v 1.24 2006/10/14 20:18:21 exodusd Exp $
 */

constant cvs_version="$Id: ldap.pike,v 1.24 2006/10/14 20:18:21 exodusd Exp $";

inherit "/kernel/module";

#include <configure.h>
#include <macros.h>
#include <config.h>
#include <attributes.h>
#include <classes.h>
#include <events.h>
#include <database.h>

//#define LDAP_DEBUG 1

#ifdef LDAP_DEBUG
#define LDAP_LOG(s, args...) werror("ldap: "+s+"\n", args)
#else
#define LDAP_LOG(s, args...)
#endif

#define LOG_AUTH(s, args...) werror("ldap-auth: "+s+"\n", args)

#define LOG_TIME( start_time, msg, args... ) { int tmp_diff_time = get_time_millis() - start_time; if ( tmp_diff_time >= log_connection_times_min && tmp_diff_time <= log_connection_times_max ) werror( "LDAP took %d milliseconds: " + msg + "\n", tmp_diff_time, args ); }


//! This module is a ldap client inside sTeam. It reads configuration
//! parameters of sTeam to contact a ldap server.
//! All the user management can be done with ldap this way. Some
//! special scenario might require to modify the modules code to 
//! make it work.
//!
//! The configuration variables used are:
//! server - the server name to connect to (if none is specified, then ldap will not be used)
//! cacheTime - how long (in seconds) shall ldap entries be cached? (to reduce server requests)
//! reuseConnection - "true" if you want to keep a single ldap connection open and reuse it, "false" if you want to close the connection after each request
//! user   - ldap user for logging in
//! password - the password for the user
//! base_dc - ldap base dc, consult ldap documentation
//! userdn - the dn path where new users are stored
//! groupdn - the dn path where new groups are stored
//! objectName - the ldap object name to be used for the search
//! userAttr - the attribute containing the user's login name
//! passwordAttr - the attribute containing the password
//! emailAttr - the attribute containing the user's email address
//! iconAttr - the attribute containing the user's icon
//! fullnameAttr - the attribute containing the user's surname
//! nameAttr - the attribute containing the user's first name
//! userClass - an attribute value that identifies users
//! userId - an attribute that contains a value that can be used to match users to groups
//! groupAttr - the attribute containing the group's name
//! groupClass - an attribute value that identifies groups
//! groupId - an attribute that contains a value that can be used to match users to groups
//! memberAttr - an attribute that contains a value that can be used to match users to groups
//! descriptionAttr - the attribute that contains a user or group description
//! notfound - defines what should be done if a user or group could not be found in LDAP:
//!            "create" : create and insert a new user in LDAP (at userdn)
//!            "ignore" : do nothing
//! sync - "true"/"false" : sync user or group data (false: read-only LDAP access)
//! bindUser - "true" : when authorizing users, try ldap bind first (you will need this if the ldap lookup doesn't return a password for the user)
//! bindUserWithoutDN - "false" : when binding users, only use the user name instead of the dn
//! requiredAttr : required attributes for creating new users in LDAP
//! adminAccount - "root" : sTeam account that will receive administrative mails about ldap
//! charset - "utf-8" : charset used in ldap server

static string sServerURL;
static mapping    config = ([ ]);

static object charset_decoder;
static object charset_encoder;

static object user_cache;
static object group_cache;
static object authorize_cache;

static object single_ldap;
static string last_bind_dn;
static string last_bind_password_hash;
static int reconnection_time = 30;

static int log_connection_times_min = Int.NATIVE_MAX;
static int log_connection_times_max = 0;

static int log_level_auth = 0;
constant LOG_AUTH_FAILURE = 1;
constant LOG_AUTH_SUCCESS = 2;
constant LOG_AUTH_DETAILS = 4;
static mapping log_failed_auth = ([ ]);

static object admin_account = 0;
static array ldap_conflicts = ({ });

static int last_error = 0;
constant NO_ERROR = 0;
constant ERROR_UNKNOWN = -1;
constant ERROR_NOT_CONNECTED = -2;


string get_identifier() { return "ldap"; }


class LDAPCache {
  static mapping by_identifier = ([ ]);
  static int cache_time;
  static function sync_func;
  static function last_modified_func;

  object create ( int cache_time_seconds, function synchronize_function,
                  function|void last_modified_function ) {
    if ( !functionp(synchronize_function) )
      FATAL( "ldap: no synchronize function specified on cache creation" );
    sync_func = synchronize_function;
    last_modified_func = last_modified_function;
    cache_time = cache_time_seconds;
  }

  array list () {
    return indices( by_identifier );
  }

  mapping find ( string identifier ) {
    return by_identifier[ identifier ];
  }

  mixed fetch ( string identifier, mixed ... args ) {
    if ( !stringp(identifier) ) return 0;
    // check whether object is already cached:
    mapping cache_entry = by_identifier[ identifier ];
    if ( mappingp(cache_entry) ) {
      if ( time() - cache_entry->last_synchronized >= cache_time ) {
        int last_modified;
        if ( !functionp(last_modified_func) ||
             (cache_entry->last_modified == 0) ||
             ((last_modified = last_modified_func( identifier, @args )) >
              cache_entry->last_modified) ) {
          mixed tmp_data = sync_func( identifier, @args );
          if ( !tmp_data ) return 0;  // only cache values != 0
          if ( stringp(tmp_data) && sizeof(tmp_data) < 1 ) return 0;
          cache_entry->data = tmp_data;
          if ( last_modified ) cache_entry->last_modified = last_modified;
          cache_entry->last_synchronized = time();
        }
      }
      cache_entry->last_accessed = time();
      return cache_entry->data;
    }
    // need to add object to cache:
    cache_entry = ([ ]);
    cache_entry->data = sync_func( identifier, @args );
    if ( !cache_entry->data ) return 0;  // only cache values != 0
    if ( stringp(cache_entry->data) && sizeof(cache_entry->data)<1 ) return 0;
    cache_entry->identifier = identifier;

    by_identifier[ identifier ] = cache_entry;
    
    if ( functionp(last_modified_func) )
      cache_entry->last_modified = last_modified_func( identifier );

    cache_entry->last_synchronized = time();
    cache_entry->last_accessed = time();

    return cache_entry->data;
  }

  void drop ( void|string identifier ) {
    if ( has_index( by_identifier, identifier ) )
      m_delete( by_identifier, identifier );
  }
}



bool ldap_activated ()
{
  if ( !stringp(config->server) || sizeof(config->server) < 1 )
    return false;
  else
    return true;
}


int get_last_error () {
  return last_error;
}


string make_dn ( mixed ... parts ) {
  string dn = "";
  foreach ( parts, mixed part ) {
    if ( stringp(part) && sizeof(part) > 0 ) {
      if ( sizeof(dn) > 0 && dn[-1] != ',' ) dn += ",";
      dn += part;
    }
  }
  return dn;
}


object connect ( void|string user, void|string password, void|bool dont_reconnect ) {
  last_error = ERROR_NOT_CONNECTED;
  if ( !ldap_activated() ) return 0;

  string bind_dn = "";
  if ( Config.bool_value( config->bindUser ) && stringp(user) ) {
    if ( Config.bool_value(config->bindUserWithoutDN) )
      bind_dn = user;
    else {
      bind_dn = make_dn( config->userAttr + "=" + user, config["userdn"], config["base_dc"] );
    }
  }
  else if ( stringp(config["user"]) && sizeof(config["user"]) > 0 ) {
    // bind with default/root user
    if ( search( config["user"], "=" ) >= 0 )
      bind_dn = config["user"];
    else
      bind_dn = make_dn( "cn="+config["user"] );
    password = config["password"];
  }
  else {  // don't bind at all:
    bind_dn = 0;
  }

  object ldap;
  mixed err = catch {
    if ( Config.bool_value(config["reuseConnection"]) ||
         config["reconnectTime"] > 0 ) {
      if ( objectp(single_ldap) && !single_ldap->is_open() )
        disconnect( single_ldap, true );
      else
        ldap = single_ldap;
    }
    else
      disconnect( single_ldap );
    if ( !objectp(ldap) ) {
      ldap = Protocols.LDAP.client( config["server"] );
      LDAP_LOG( "new connection: %O", ldap );
    }
  };
  if ( err ) werror( "ldap: error while connecting: %O\n", err[0] );
  if ( !objectp(ldap) ) return 0;

  single_ldap = ldap;

  err = catch {
    if ( stringp(bind_dn) && stringp(password) ) {
      if ( bind_dn != last_bind_dn ||
           !verify_crypt_md5( password, last_bind_password_hash ) ) {
        LDAP_LOG( "binding %O", bind_dn );
        if ( !ldap->bind( bind_dn, password ) )
          throw( "bind failed on " + bind_dn );
        // remember dn and password hash, so we don't need to re-bind if
        // the user and password haven't changed:
        last_bind_dn = bind_dn;
        last_bind_password_hash = make_crypt_md5( password );
      }
    }
    else LDAP_LOG( "using without bind" );
    ldap->set_scope( 2 );
    if ( stringp(config["base_dc"]) && sizeof(config["base_dc"]) )
      ldap->set_basedn( config["base_dc"] );
  };
  if ( err != 0 ) {
    string error_msg = "";
    mixed ldap_error_nr = ldap->error_number();
    if ( ldap_error_nr > 0 )
      error_msg = "\n" + ldap->error_string() +
	sprintf( " (#%d)", ldap_error_nr );
    LDAP_LOG( "Failed to bind " + bind_dn + " on ldap" + error_msg );
    if ( ldap_error_nr == 0x31 || // invalid credentials
         ldap_error_nr == 0x30 ) {  // no such object
      disconnect( ldap );
      return 0;
    }
    else {
      disconnect( ldap, true );
      last_error = ERROR_NOT_CONNECTED;
      if ( dont_reconnect ) {
        FATAL("Failed to bind " + bind_dn + " on ldap" + error_msg);
        return 0;
      }
      else
        return connect( user, password, true );  // try to reconnect once
    }
  }
  last_error = NO_ERROR;
  return ldap;
}


void disconnect ( object ldap, void|bool force_disconnect ) {
  last_error = NO_ERROR;
  if ( !objectp(ldap) ) return;
  if ( Config.bool_value(config["reuseConnection"]) ||
       config["reconnectTime"] > 0 ) {
    if ( !force_disconnect ) return;
    single_ldap = 0;
  }
  last_bind_dn = 0;
  last_bind_password_hash = 0;
  int fd = ldap->query_fd();
  destruct( ldap );
  if ( fd > 600 ) {
    werror( "ldap: %d open file descriptors, calling garbage collector...\n",
            sizeof( Stdio.get_all_active_fd() ) );
    int time_start = get_time_millis();
    gc();  // force garbage collection to free ldap file descriptors
    werror( "ldap: garbage collector done, %d open file descriptors remain.\n",
            sizeof( Stdio.get_all_active_fd() ) );
    LOG_TIME( time_start, "disconnect[gc]" );
  }
  LDAP_LOG( "connection closed (file descriptor was %d)", fd );
}


mapping get_config ()
{
  return config;
}


object get_user_cache () {
  if ( !GROUP("admin")->is_member( this_user() ) ) return 0;
  return user_cache;
}


object get_group_cache () {
  if ( !GROUP("admin")->is_member( this_user() ) ) return 0;
  return group_cache;
}


object get_authorize_cache () {
  if ( !GROUP("admin")->is_member( this_user() ) ) return 0;
  return authorize_cache;
}


mapping get_failed_authorize () {
  if ( !GROUP("admin")->is_member( this_user() ) ) return 0;
  return log_failed_auth;
}


static void init_module()
{
    last_error = NO_ERROR;
    config = Config.read_config_file( _Server.get_config_dir()+"/modules/ldap.cfg", "ldap" );
    if ( !mappingp(config) ) {
        config = ([ ]);
	MESSAGE("LDAP Service not started - missing configuration !");
        last_error = ERROR_NOT_CONNECTED;
	return; // ldap not started !
    }
    if ( !ldap_activated() ) {
      MESSAGE("LDAP deactivated.");
      last_error = ERROR_NOT_CONNECTED;
      return;  // ldap deactivated
    }

    LDAP_LOG("configuration is %O", config);
    if ( stringp(config["charset"]) ) {
      charset_decoder = Locale.Charset.decoder( config["charset"] );
      if ( !objectp(charset_decoder) )
        werror( "LDAP: could not create a charset converter for %s\n",
                config["charset"] );
    }
    else charset_decoder = 0;
    charset_encoder = Locale.Charset.encoder( "utf-8" );

    int cache_time = config["cacheTime"];
    user_cache = LDAPCache( cache_time, fetch_user_internal,
                            user_last_modified );
    group_cache = LDAPCache( cache_time, fetch_group_internal,
                             group_last_modified );
    authorize_cache = LDAPCache( cache_time, authorize_ldap_internal,
                                 user_last_modified );

    if ( stringp(config->notfound) && lower_case(config->notfound) == "create"
         && !config->objectName )
      steam_error("objectName configuration missing !");

    if ( Config.bool_value( config->sync ) ) {
      // if our main dc does not exist - create it
      object ldap = connect();
      mixed err = catch {
        ldap->add( config->base_dc, ([
          "objectclass": ({ "dcObject", "organization" }),
       	  "o": "sTeam Authorization Directory",
          "dc": "steam" ]) );
      };
      if ( err ) werror( "ldap: failed to create main dc: %O\n", err[0] );
      disconnect( ldap );
    }
    else {
      // check whether the connection is working
      object ldap;
      mixed err = catch( ldap = Protocols.LDAP.client( config["server"] ) );
      if ( err ) werror( "ldap: error while connecting: %O\n", err[0] );
      if ( objectp(ldap) ) {
	MESSAGE( "LDAP: connected to %s", config["server"] );
	disconnect( ldap );
      }
      else {
        last_error = ERROR_NOT_CONNECTED;
	MESSAGE( "LDAP: failed to connect to %s", config["server"] );
	werror( "LDAP: failed to connect to %s\n", config["server"] );
      }
    }
}


void load_module()
{
  add_global_event(EVENT_USER_CHANGE_PW, sync_password, PHASE_NOTIFY);
  add_global_event(EVENT_USER_NEW_TICKET, sync_ticket, PHASE_NOTIFY);
}


private static bool notify_admin ( string msg ) {
  if ( zero_type( config["adminAccount"] ) ) return false;
  object admin = USER( config["adminAccount"] );
  if ( !objectp(admin) ) admin = GROUP( config["adminAccount"] );
  if ( !objectp(admin) ) return false;
  string msg_text = "The following LDAP situation occured on the server "
    + _Server->get_server_name() +" at "+ (ctime(time())-"\n") + " :\n" + msg;
  admin->mail( msg_text, "LDAP on " + _Server->get_server_name(), 0, "text/plain" );
  return true;
}


static mixed map_results(object results)
{
  array result = ({ });
  
  for ( int i = 1; i <= results->num_entries(); i++ ) {
    mapping            data = ([ ]);
    mapping res = results->fetch(i);
    
    foreach(indices(res), string attr) {
      if ( arrayp(res[attr]) ) {
	if ( sizeof(res[attr]) == 1 )
	  data[attr] = res[attr][0];
	else
	  data[attr] = res[attr];
      }
    }
    if ( results->num_entries() == 1 )
      return data;
    result += ({ data });
  }
  return result;
}


void uncache_user ( string user ) {
  user_cache->drop( user );
  authorize_cache->drop( user );
}


void uncache_group ( string group ) {
  group_cache->drop( group );
}


mapping search_user ( string search_str, void|string user, void|string pass )
{
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return UNDEFINED;
  }

  if ( !stringp(search_str) || sizeof(search_str)<1 ) {
    werror( "LDAP: search_user: invalid search_str: %O\n", search_str );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  mapping udata = ([ ]);
  object results;
  
  int time_start = get_time_millis();
  
  object ldap;
  if ( Config.bool_value(config->bindUser) && stringp(user) && stringp(pass) )
    ldap = connect(user, pass);
  else
    ldap = connect();

  LDAP_LOG("searching %s (user %O) : ldap is %O", search_str, user, ldap);

  if ( !objectp(ldap) ) {
    last_error = ERROR_NOT_CONNECTED;
    return UNDEFINED;
  }

  LDAP_LOG("looking up user in LDAP: %s\n", search_str);
  if ( config->userdn ) {
    ldap->set_basedn( make_dn( config->userdn, config->base_dc ) );
    if ( ldap->error_number() )
      werror( "ldap: error searching user: %s (#%d)\n",
	      ldap->error_string(), ldap->error_number() );
  }

  mixed err = catch( results = ldap->search( search_str ) );
  if ( err ) {
    werror( "ldap: error searching user: %s\n", err[0] );
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  LOG_TIME( time_start, "search_user( %s )", search_str );

  if ( objectp(ldap) && ldap->error_number() ) {
    werror( "ldap: Error while searching user: %s (#%d)\n",
	    ldap->error_string(), ldap->error_number() );
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  if ( !objectp(results) ) {
    werror("ldap: Invalid results while searching user.\n");
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }
    
  if ( results->num_entries() == 0 ) {
    LDAP_LOG("User not found in LDAP directory: %s", search_str);
    disconnect( ldap );
    return UNDEFINED;
  }

  udata = map_results( results );
  LDAP_LOG( "user %s has dn: %O and %d data entries",
            search_str, udata["dn"], sizeof(indices(udata)) );

  disconnect( ldap );

  LOG_TIME( time_start, "search_user( %s )", search_str );

  return udata;
}


mapping search_group ( string search_str )
{
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return UNDEFINED;
  }

  if ( !stringp(search_str) || sizeof(search_str)<1 ) {
    werror( "LDAP: search_group: invalid search_str: %O\n", search_str );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }
  
  mapping gdata = ([ ]);
  object results;

  if ( !stringp(config->groupAttr) || sizeof(config->groupAttr)<1 ) {
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  int time_start = get_time_millis();

  object ldap = connect();

  if ( !objectp(ldap) ) {
    last_error = ERROR_NOT_CONNECTED;
    return UNDEFINED;
  }

  LDAP_LOG("looking up group in LDAP: %s\n", search_str);
  if ( config->groupdn ) {
    ldap->set_basedn( make_dn( config->groupdn, config->base_dc ) );
    if ( ldap->error_number() )
      werror( "ldap: error searching group: %s (#%d)\n",
	      ldap->error_string(), ldap->error_number() );
  }

  mixed err = catch( results = ldap->search( search_str ) );
  if ( err ) {
    werror( "ldap: error searching group: %s\n", err[0] );
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  LOG_TIME( time_start, "search_group( %s )", search_str );

  if ( objectp(ldap) && ldap->error_number() ) {
    werror( "ldap: Error while searching group: %s (#%d)\n",
	    ldap->error_string(), ldap->error_number() );
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  if ( !objectp(results) ) {
    werror("ldap: Invalid results while searching group.\n");
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }
    
  if ( results->num_entries() == 0 ) {
    LDAP_LOG("Group not found in LDAP directory: %s", search_str);
    disconnect( ldap );
    return UNDEFINED;
  }

  gdata = map_results(results);
  LDAP_LOG( "group %s has dn: %O", search_str, gdata["dn"] );

  disconnect( ldap );

  return gdata;
}


int user_last_modified ( string identifier, void|string pass )
{
  //TODO: query ldap attribute modifyTimestamp
  return 0;
}


mapping fetch_user ( string identifier, void|string pass )
{
  int time_start = get_time_millis();
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return 0;
  }
  if ( !stringp(identifier) ) {
    last_error = ERROR_UNKNOWN;
    return 0;
  }

  LDAP_LOG("fetch_user(%s) %d", identifier, stringp(pass));

  mapping result = user_cache->fetch( identifier, pass );
  if ( !mappingp(result) ) authorize_cache->drop( identifier );
  LOG_TIME( time_start, "fetch_user( %s )", identifier );
  return result;
}


mapping fetch_user_internal ( string identifier, void|string pass )
{
  LDAP_LOG("fetch_user_internal(%s) %d", identifier, stringp(pass));
  string search_str = "("+config->userAttr+"="+identifier+")";
  if ( stringp(config->userClass) && sizeof(config->userClass)>0 )
    search_str = "(&"+search_str+"(objectclass="+config->userClass+"))";
  return fix_charset( search_user( search_str, identifier, pass ) );
}


int group_last_modified ( string identifier )
{
  //TODO: query ldap attribute modifyTimestamp
  return 0;
}


mapping fetch_group ( string identifier )
{
  int time_start = get_time_millis();
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return 0;
  }
  if ( !stringp(identifier) ) {
    last_error = ERROR_UNKNOWN;
    return 0;
  }

  if ( !stringp(config->groupAttr) || sizeof(config->groupAttr)<1 ) {
    last_error = ERROR_UNKNOWN;
    return 0;
  }

  LDAP_LOG("fetch_group(%s)", identifier);
  mapping result = group_cache->fetch( identifier );
  LOG_TIME( time_start, "fetch_group( %s )", identifier );
  return result;
}


mapping fetch_group_internal ( string identifier )
{
  LDAP_LOG("fetch_group_internal(%s)", identifier);
  string search_str = "("+config->groupAttr+"="+identifier+")";
  if ( stringp(config->groupClass) && sizeof(config->groupClass)>0 )
    search_str = "(&"+search_str+"(objectclass="+config->groupClass+"))";

  return fix_charset( search_group( search_str ) );
}


mapping fetch ( string dn, string pattern )
{
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return 0;
  }
  if ( !stringp(dn) || !stringp(pattern) ) {
    last_error = ERROR_UNKNOWN;
    return 0;
  }

  // caller must be module...
  if ( !_Server->is_module( CALLER ) )
    steam_error( "Access for non-module denied !" );

  object results;

  object ldap = connect();

  if ( !objectp(ldap) ) {
    last_error = ERROR_NOT_CONNECTED;
    return UNDEFINED;
  }

  if ( stringp(dn) && sizeof(dn)>0 )
    ldap->set_basedn( make_dn( dn, config->base_dc ) );
  else
    ldap->set_basedn( config->base_dc );
  if ( ldap->error_number() )
    werror( "ldap: error fetching: %s (#%d)\n",
	    ldap->error_string(), ldap->error_number() );

  mixed err = catch( results = ldap->search( pattern ) );
  if ( err ) {
    werror( "ldap: error fetching: %s\n", err[0] );
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }
  
  if ( objectp(ldap) && ldap->error_number() ) {
    werror( "ldap: Error while fetching: %s (#%d)\n",
	    ldap->error_string(), ldap->error_number() );
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }

  if ( !objectp(results) ) {
    werror("ldap: Invalid results while fetching.\n");
    disconnect( ldap );
    last_error = ERROR_UNKNOWN;
    return UNDEFINED;
  }
    
  if ( results->num_entries() == 0 ) {
    LDAP_LOG("No results when fetching: %s", pattern);
    disconnect( ldap );
    return UNDEFINED;
  }

  disconnect( ldap );

  return map_results( results );
}


mixed fix_charset ( string|mapping|array v )
{
  if ( zero_type(v) ) return UNDEFINED;
  if ( !objectp(charset_encoder) || !objectp(charset_decoder) ) return v;
  if ( stringp(v) ) {
    if ( xml.utf8_check(v) ) return v;  // already utf-8
    string tmp = charset_decoder->clear()->feed(v)->drain();
    tmp = charset_encoder->clear()->feed(tmp)->drain();
    LDAP_LOG( "charset conversion: from \"%s\" to \"%s\".", v, tmp );
    return tmp;
  }
  else if ( arrayp(v) ) {
    array tmp = ({ });
    foreach ( v, mixed i )
      tmp += ({ fix_charset(i) });
    return tmp;
  }
  else if ( mappingp(v) ) {
    mapping tmp = ([ ]);
    foreach ( indices(v), mixed i )
      tmp += ([ fix_charset(i) : fix_charset(v[i]) ]);
    return tmp;
  }
  else return UNDEFINED;
}


static bool check_password(string pass, string user_pw)
{
  if ( !stringp(pass) || !stringp(user_pw) )
    return false;

  LDAP_LOG("check_password()");
  if ( sizeof(user_pw) > 5 && user_pw[0..4] == "{SHA}" )
    return user_pw[5..] == MIME.encode_base64( sha_hash(pass) );
  if ( sizeof(user_pw) > 7 && user_pw[0..6] == "{crypt}" )
    return crypt(pass, user_pw[7..]);
  return verify_crypt_md5(pass, user_pw);
}


bool authorize_ldap ( object user, string pass )
{
  int time_start = get_time_millis();
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return false;
  }

  if ( !objectp(user) )
    steam_error("User object expected for authorization !");

  if ( !stringp(pass) || sizeof(pass)<1 ) {
    last_error = ERROR_UNKNOWN;
    return false;
  }

  string uname = user->get_user_name();

  // don't authorize restricted users:
  if ( _Persistence->user_restricted( uname ) ) return false;

  string cached = authorize_cache->fetch( uname, pass );

  if ( stringp(cached) && check_password( pass, cached ) ) {
    // successfully authorized from cache
    LDAP_LOG("user %s LDAP cache authorized", uname);
    LOG_TIME( time_start, "authorize_ldap[success]( %s )", uname );
    bool log_auth = false;
    if ( log_level_auth & LOG_AUTH_FAILURE ) {
      if ( has_index( log_failed_auth, uname ) ) {
        LOG_AUTH( "%s finally authorized by ldap cache after %d seconds "+
          "(crc: %d)", uname, time()-log_failed_auth[uname], Gz.crc32(pass) );
        m_delete( log_failed_auth, uname );
        log_auth = true;
      }
    }
    if ( log_level_auth & LOG_AUTH_SUCCESS ) {
      LOG_AUTH( "%s authorized by ldap cache (crc: %d)",
                uname, Gz.crc32(pass) );
      log_auth = true;
    }
    return true;
  }

  // failed to authorize from cache
  if ( log_level_auth & LOG_AUTH_FAILURE ) {
    if ( !has_index( log_failed_auth, uname ) ) {
      log_failed_auth[ uname ] = time();
      LOG_AUTH( "%s : %s failed", (ctime(time())-"\n"), uname );
    }
    LOG_AUTH( "%s not authorized by ldap cache (password check) (crc: %d)",
            uname, Gz.crc32(pass) );
  }
  LDAP_LOG("user %s found in LDAP cache - password failed: %O",
           uname, cached);
  LOG_TIME( time_start, "authorize_ldap[failed]( %s )", uname );
  return false;
}


static string authorize_ldap_internal ( string uname, string pass )
{
  int time_start = get_time_millis();

  // try authorizing via bind:
  if ( Config.bool_value(config->bindUser) &&
       stringp(uname) && sizeof(uname) > 0 &&
       stringp(pass) && sizeof(pass) > 0) {
    object ldap = connect( uname, pass );
    if ( objectp(ldap) ) {
      disconnect( ldap );
      LDAP_LOG("authorized %s via bind", uname);
      LOG_TIME( time_start, "authorize_ldap[bind]( %s )", uname );
      string cached_pw = make_crypt_md5( pass );
      bool log_auth = false;
      if ( (log_level_auth & LOG_AUTH_FAILURE) &&
           has_index( log_failed_auth, uname ) ) {
        LOG_AUTH( "%s finally authorized via ldap bind (after %d "+
                "seconds)", uname, time() - log_failed_auth[uname] );
        log_auth = true;
      }
      if ( log_level_auth & LOG_AUTH_SUCCESS ) {
        LOG_AUTH( "%s authorized via ldap bind", uname );
        log_auth = true;
      }
      if ( log_auth && (log_level_auth & LOG_AUTH_DETAILS) )
        LOG_AUTH( "%s password crc32 : %d", uname, Gz.crc32( pass ) );
      return cached_pw;
    }
    else {
      LDAP_LOG("could not authorize %s via bind (no ldap connection)", uname);
      if ( log_level_auth & LOG_AUTH_FAILURE ) {
        LOG_AUTH( "%s not authorized via ldap bind", uname );
        if ( log_level_auth & LOG_AUTH_DETAILS )
          LOG_AUTH( "%s password crc32: %d", uname, Gz.crc32( pass ) );
      }
    }
  }

  // fetch user data and authorize via password:
  mapping udata = fetch_user(uname, pass);
  LDAP_LOG("trying to authorize user %s via password", uname);

  LOG_TIME( time_start, "authorize_ldap[fetch]( %s )", uname );

  if ( mappingp(udata) ) {
    object user = USER( uname );
    string dn = udata["dn"];
    if ( !stringp(dn) ) dn = "";

    // check for conflicts (different user in sTeam than in LDAP):
    if (config->checkConflicts && !stringp(user->query_attribute("ldap:dn"))) {
      if ( search(ldap_conflicts,uname)<0 ) {
	ldap_conflicts += ({ uname });
        if ( notify_admin(
	    "Dear LDAP administrator at "+_Server->get_server_name()
	    +",\n\nthere has been a conflict between LDAP and sTeam:\n"
	    +"User \""+uname+"\" already exists in sTeam, but now "
	    +"there is also an LDAP user with the same name/id.\nYou "
	    +"will need to remove/rename one of them or, if they are "
	    +"the same user, you can overwrite the sTeam data from LDAP "
	    +"by adding a \"dn\" attribute to the sTeam user." ) );
	else
	  werror( "ldap: user conflict: %s in sTeam vs. %s in LDAP\n", uname, dn );
	return 0;
      }
    }
    else if ( search(ldap_conflicts,uname) >= 0 )
      ldap_conflicts -= ({ uname });

    if ( check_password( pass, udata[config->passwordAttr] ) ) {
      // need to synchronize passwords from ldap if ldap is down ?!
      // this is only done when the ldap password is received
      if ( udata[config->passwordAttr] != user->get_user_password())
	user->set_user_password(udata[config->passwordAttr], 1);
      LDAP_LOG("user %s authorized via password from LDAP", uname);
      string cached_pw = make_crypt_md5( pass );
      bool log_auth = false;
      if ( (log_level_auth & LOG_AUTH_FAILURE) &&
           has_index( log_failed_auth, uname ) ) {
        LOG_AUTH( "%s finally authorized via password (after %d seconds)",
                  uname, time() - log_failed_auth[uname] );
        log_auth = true;
      }
      if ( log_level_auth & LOG_AUTH_SUCCESS ) {
        LOG_AUTH( "%s authorized via password", uname );
        log_auth = true;
      }
      if ( log_auth && (log_level_auth & LOG_AUTH_DETAILS) )
        LOG_AUTH( "%s password crc32: %d\n* ldap hash: %O",
                uname, Gz.crc32(pass), udata[config->passwordAttr] );
      return cached_pw;
    }
    else {
      if ( log_level_auth & LOG_AUTH_FAILURE ) {
        LOG_AUTH( "%s not authorized via password", uname );
        if ( log_level_auth & LOG_AUTH_DETAILS )
          LOG_AUTH( "%s password crc32: %d\n* ldap hash: %O",
                    uname, Gz.crc32(pass), udata[config->passwordAttr] );
      }
      LDAP_LOG("user %s found in LDAP directory - password failed!", uname);
      return 0;
    }
  }
  LDAP_LOG("user " + uname + " was not found in LDAP directory.");
  // if notfound configuration is set to create, then we should create a user:
  if ( config->notfound == "create" ) {
    object user = USER( uname );
    if ( add_user( uname, pass, user ) )
      return make_crypt_md5( pass );
  }
  return 0;
}

object sync_user(string name)
{
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return 0;
  }

  // don't sync restricted users:
  if ( _Persistence->user_restricted( name ) )
    return 0;

  mapping udata = fetch_user( name );
  if ( !mappingp(udata) )
    return 0;

  LDAP_LOG("sync of ldap user \"%s\": %O", name, udata);
  object user = get_module("users")->get_value(name);
  if ( objectp(user) ) {
    // update user date from LDAP
    if ( ! user->set_attributes( ([
             "pw" : udata[config->passwordAttr],
	     "email" : udata[config->emailAttr],
	     "fullname" : udata[config->fullnameAttr],
	     "firstname" : udata[config->nameAttr],
	     "OBJ_DESC" : udata[config->descriptionAttr],
	   ]) ) )
      werror( "LDAP: Could not sync user attributes with ldap for \"%s\".\n", name );
  } else {
    // create new user to match LDAP user
    object factory = get_factory(CLASS_USER);
    user = factory->execute( ([
	     "name" : name,
	     "pw" : udata[config->passwordAttr],
	     "email" : udata[config->emailAttr],
	     "fullname" : udata[config->fullnameAttr],
	     "firstname" : udata[config->nameAttr],
	     "OBJ_DESC" : udata[config->descriptionAttr],
	   ]) );
    user->set_user_password( udata[config->passwordAttr], 1 );
    user->activate_user( factory->get_activation() );
  }
  // sync group membership:
  if ( objectp( user ) ) {
    string primaryGroupId = udata[config->groupId];
    if ( stringp( primaryGroupId ) ) {
      mapping group = search_group("("+config->groupId+"="+primaryGroupId+")");
    }
  }

  return user;
}

object sync_group(string name)
{
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return 0;
  }

  // don't syncronize restricted groups:
  if ( _Persistence->group_restricted( name ) )
    return 0;

  mapping gdata = fetch_group( name );
  if ( !mappingp(gdata) )
    return 0;

  LDAP_LOG("sync of ldap group: %O", gdata);
  //object group = get_module("groups")->lookup(name);
  object group = get_module("groups")->get_value( name );
  if ( objectp(group) ) {
    // group memberships are handled by the persistence manager
    // update group date from LDAP
    group->set_attributes( ([
	     "OBJ_DESC" : gdata[config->descriptionAttr],
	   ]) );
  } else {
    // create new group to match LDAP group
    object factory = get_factory(CLASS_GROUP);
    group = factory->execute( ([
	      "name": name,
	      "OBJ_DESC" : gdata[config->descriptionAttr],
	    ]) );
  }
  return group;
}

static void sync_password(int event, object user, object caller)
{
  if ( !Config.bool_value(config->sync) )
    return;
  string oldpw = user->get_old_password();
  string crypted = user->get_user_password();
  string name = user->get_user_name();
  // don't sync password for restricted users:
  if ( _Persistence->group_restricted( name ) ) return;
  LDAP_LOG("password sync for " + user->get_user_name());

  object ldap;
  string dn;

  if ( Config.bool_value(config->bindUser) && oldpw &&
       objectp(ldap = connect( name, oldpw )) ) {
    if ( config->userdn )
      dn = make_dn(config->userAttr+"="+name, config->userdn, config->base_dc);
    else
      dn = make_dn(config->userAttr+"="+name, config->base_dc);
  }
  else if ( ldap = connect() ) {
    dn = make_dn( config->base_dc, config->userAttr+"="+name );
  }

  if ( !stringp(dn) ) {
    LDAP_LOG("sync_password(): no dn");
    disconnect( ldap );
    return;
  }

  mixed err;
  if ( crypted[..2] == "$1$" )
    crypted = "{crypt}" + crypted;
  err = catch( ldap->modify(dn, ([ config->passwordAttr: ({ 2,crypted }),])) );
  authorize_cache->drop( name );
  user_cache->drop( name );
  LDAP_LOG("sync_password(): %s - %s - %O\n", crypted, dn, ldap->error_string());
  disconnect( ldap );
}

static void sync_ticket(int event, object user, object caller, string ticket)
{
  last_error = NO_ERROR;
  mixed err;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return;
  }
  if ( !Config.bool_value(config->sync) )
    return;
  string name = user->get_user_name();
  string dn = make_dn( config->base_dc, config->userAttr + "=" + name );
  object ldap = connect();
  err = catch( ldap->modify(dn, ([ "userCertificate": ({ 2,ticket }),])) );
  disconnect( ldap );
}

bool is_user(string user)
{
  last_error = NO_ERROR;
  object ldap = connect();
  if ( !objectp(ldap) ) {
    last_error = ERROR_NOT_CONNECTED;
    return false;
  }
  object results = ldap->search( "("+config["userAttr"]+"="+user+")" );
  bool result = (objectp(results) && results->num_entries() > 0);
  disconnect( ldap );
  return result;
}

static bool add_user(string name, string password, object user)
{
  last_error = NO_ERROR;
  if ( !ldap_activated() ) {
    last_error = ERROR_NOT_CONNECTED;
    return false;
  }
  if ( !Config.bool_value(config->sync) ) return false;
  if ( !stringp(config->notfound) || lower_case(config->notfound) != "create" )
    return false;

  // don't add restricted users:
  if ( _Persistence->user_restricted( name ) ) return false;
  
  string fullname = user->get_name();
  string firstname = user->query_attribute(USER_FIRSTNAME);
  string email = user->query_attribute(USER_EMAIL);
  
  mapping attributes = ([
    config["userAttr"]: ({ name }),
    config["fullnameAttr"]: ({ fullname }),
    "objectClass": ({ config["objectName"] }),
    config["passwordAttr"]: ({ make_crypt_md5(password) }),
  ]);
  if ( stringp(firstname) && strlen(firstname) > 0 )
    config["nameAttr"] = ({ firstname });
  if ( stringp(email) && strlen(email) > 0 )
    config["emailAttr"] = ({ email });

  array(string) requiredAttributes =  config["requiredAttr"];
  
  if ( arrayp(requiredAttributes) && sizeof(requiredAttributes) > 0 ) {
    foreach(requiredAttributes, string attr) {
      if ( zero_type(attributes[attr]) )
	attributes[attr] = ({ "-" });
    }
  }
  
  object ldap = connect();
  if ( !objectp(ldap) ) {
    last_error = ERROR_NOT_CONNECTED;
    return false;
  }
  ldap->add( make_dn(config["userAttr"]+"="+name, config["base_dc"]),
             attributes );
  int err = ldap->error_number();
  if ( err != 0 )
    FATAL( "Failed to add user , error is " + ldap->error_string() );
  bool result = ldap->error_number() == 0;
  disconnect( ldap );
  return result;
}


/**
 * Returns the minimum number of milliseconds a connection request must take
 * to appear in the log. Note: if logging is switched off, then this will be
 * the maximum value an integer can take.
 *
 * @return the minimum threshold (in milliseconds) for connection times logging
 */
int get_log_connection_times_min () {
  return log_connection_times_min;
}


/**
 * Returns the maximum number of milliseconds a connection request must take
 * to appear in the log. Note: if no upper threshold is used, then this will be
 * the maximum value an integer can take.
 *
 * @return the maximum threshold (in milliseconds) for connection times logging
 */
int get_log_connection_times_max () {
  return log_connection_times_max;
}


/**
 * Activate or deactivate connection times logging and set the threshold for
 * log entries.
 *
 * @param min_milliseconds the minimum number of milliseconds a connection
 *   request must take to appear in the log (if -1, then logging will be
 *   deaktivated)
 * @param max_milliseconds the maximum number of milliseconds a connection
 *   request must take to appear in the log (if -1, then no upper limit will
 *   be used)
 * @return true if logging is now active, false if it is now not active
 */
bool set_log_connection_times ( int min_milliseconds, void|int max_milliseconds ) {
  if ( min_milliseconds < 0 )
    log_connection_times_min = Int.NATIVE_MAX;
  else
    log_connection_times_min = min_milliseconds;
  if ( zero_type(max_milliseconds) || max_milliseconds < 0 )
    log_connection_times_max = Int.NATIVE_MAX;
  else
    log_connection_times_max = max_milliseconds;
  return log_connection_times_min < Int.NATIVE_MAX &&
    log_connection_times_max >= log_connection_times_min;
}


static int get_time_millis () {
  array tod = System.gettimeofday();
  return tod[0]*1000 + tod[1]/1000;
}


/**
 * Set the log level for authentication logging during runtime.
 * The log levels are:
 * 0 : no logging,
 * 1 : failed authentication,
 * 2 : successful authentication,
 * 3 : failed and successful authentication,
 * 5 : detailed failed authentication (with password hashes in the log),
 * 6 : detailed successful authentication (with password hashes in the log),
 * 7 : detailed failed and successful authentication (with password hashes in
 *     the log).
 *
 * @param log_level authentication log level
 * @return the new log level
 */
int set_authorize_log_level ( int log_level ) {
  log_level_auth = log_level;
  if ( log_level_auth < 1 ) log_failed_auth = ([ ]);
  return log_level_auth;
}


/**
 * Query the authentication log level
 * The log levels are:
 * 0 : no logging,
 * 1 : failed authentication,
 * 2 : successful authentication,
 * 3 : failed and successful authentication,
 * 5 : detailed failed authentication (with password hashes in the log),
 * 6 : detailed successful authentication (with password hashes in the log),
 * 7 : detailed failed and successful authentication (with password hashes in
 *     the log).
 *
 * @return the authentication log level
 */
int get_authorize_log_level () {
  return log_level_auth;
}


/*
void synchronization_thread () {
  while ( 1 ) {
    int cache_time = config["cacheTime"];
    int cache_live_time = config["cacheTime"] * 2;  //TODO: make this configurable
    foreach ( indices( user_cache ), string identifier ) {
      mapping entry = user_cache[ identifier ];
      int t = time();
      if ( t - entry->last_accessed > cache_live_time ) {
        //TODO: drop entry
        continue;
      }
      if ( t - entry->last_synchronized > cache_time ) {
        //TODO: sync entry
      }
    }
    foreach ( indices( group_cache ), string identifier ) {
    }
    foreach ( indices( authorize_cache ), string identifier ) {
    }
    sleep( cacheTime ); //TODO: substract the time needed for sync
  }
}
*/
